/*
 * (C) Copyright 2012 Henrik Nordstrom <henrik@henriknordstrom.net>
 * (C) Copyright 2012 Alejandro Mery <amery@geeks.cl>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#ifndef SUNXI_TYPES_H
#define SUNXI_TYPES_H

#include <stdint.h>

#define __s8	int8_t
#define __s16	int16_t
#define __s32	int32_t
#define __s64	int64_t

#define s8	int8_t
#define s16	int16_t
#define s32	int32_t
#define s64	int64_t

#define __u8	uint8_t
#define __u16	uint16_t
#define __u32	uint32_t
#define __u64	uint64_t

#define u8	uint8_t
#define u16	uint16_t
#define u32	uint32_t
#define u64	uint64_t

#endif
